package com.bernhardriedl.projects.gameoflife.logic;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.StringTokenizer;
import java.util.Vector;

import com.bernhardriedl.projects.gameoflife.examples.ExamplesContainer;
import com.bernhardriedl.projects.tools.utils.ReadFile;
import com.bernhardriedl.projects.tools.utils.WriteFile;

public class CreaturesMatrix {

	private final int beBorn=3;
	private final int keepAlive_min=2;
	private final int keepAlive_max=3;

	public final static int randomLimit=100;

	private boolean [][] data=null;
	private boolean hasBounds=true;

	/**
	 * Constructor - Random Universe
	 * @param x width
	 * @param y height
	 * @param lifeFactor probability for life
	 * @throws Exception
	 */

	public CreaturesMatrix(int x, int y, int lifeFactor) throws Exception {
		this.buildFromRandom(x, y, lifeFactor);
	}

	/**
	 * Constructor - Empty Universe
	 * @param x width
	 * @param y height
	 * @throws Exception
	 */

	public CreaturesMatrix(int x, int y) throws Exception {
		this.buildEmpty(x, y);
	}

	/**
	 * Constructor - Universe from File
	 * @param fileName source-file
	 * @throws Exception
	 */

	public CreaturesMatrix(String fileName) throws Exception {
		this.buildFromFile(fileName);
	}

	/**
	 * setter LifeFactor
	 * @param lifeFactor probability for life
	 * @throws Exception
	 */

	private void controlLifeFactor(int lifeFactor) throws Exception {
		if (lifeFactor<0 || lifeFactor>randomLimit)
			throw new Exception ("lifeFactor (k) out of bounds! 0 <= k <= "+randomLimit);
	}

	/**
	 * controlSize
	 * @param value width/height
	 * @throws Exception
	 */

	private void controlSize(int value) throws Exception {
		if (value<3)
			throw new Exception("dimension can't be smaller than 3!");
	}

	/**
	 * initialize Universe
	 * @param x width
	 * @param y height
	 * @throws Exception
	 */

	public void buildMatrix(int x, int y) throws Exception {
		this.controlSize(x);
		this.controlSize(y);

		this.data=new boolean [x][y];
	}

	/**
	 * build empty matrix
	 * @param x width
	 * @param y height
	 * @throws Exception
	 */

	public void buildEmpty(int x, int y) throws Exception {
		this.buildMatrix(x, y);

		for (int i=0; i<x; i++) {
			for (int j=0; j<y; j++) {
				this.data[i][j]=false;
			}
		}
	}

	/**
	 * build random Matrix
	 * @param x width
	 * @param y height
	 * @param lifeFactor probability for life
	 * @throws Exception
	 */

	public void buildFromRandom(int x, int y, int lifeFactor) throws Exception {
		this.controlLifeFactor(lifeFactor);

		this.buildMatrix(x, y);

		for (int i=0; i<this.data.length; i++) {
			for (int j=0; j<this.data[i].length; j++) {
				int probability=com.bernhardriedl.projects.tools.utils.Utils.getLimitedRandom(CreaturesMatrix.randomLimit);

				if (lifeFactor>=probability)
					this.data[i][j]=true;
				else
					this.data[i][j]=false;
			}
		}
	}

	/**
	 * restore Matrix from Stream
	 * @param name Filename
	 * @throws Exception
	 */

	public void buildFromStream(String name) throws Exception {
		InputStream is = null;
	    BufferedReader br = null;
	    
	    is=ExamplesContainer.class.getResourceAsStream(name+".gol");
	    br = new BufferedReader(new InputStreamReader(is));
	    
	    Vector <String> lines=new Vector <String>();
	    String line=null;
	    
	    while ((line = br.readLine()) != null)
	         lines.add(line);
	    
	    this.buildFromVector(lines);

		br.close();
		is.close();
	}

	/**
	 * restore Matrix from File
	 * @param fileName Filename
	 * @throws Exception
	 */

	public void buildFromFile(String fileName) throws Exception {
		ReadFile readFile=new ReadFile(fileName);
		readFile.open();

		Vector <String> lines=readFile.readLines();
		
		readFile.close();

		this.buildFromVector(lines);	
	}
	
	/**
	 * restore Matrix from Vector
	 * @param lines GameOfLife Universe Vector
	 * @throws Exception
	 */
	
	public void buildFromVector(Vector <String> lines) throws Exception {
		
	    /**
		 * read Dimension
		 */

		int x=Integer.parseInt(lines.elementAt(0));
		int y=Integer.parseInt(lines.elementAt(1));

		this.buildMatrix(x, y);
		
		/**
		 * read all Lines
		 */

		for (int i=0; i<x; i++) {
			String line=lines.elementAt(i+2);
			StringTokenizer tokens=new StringTokenizer(line);

			for (int j=0; j<y; j++) {
				Boolean bol=new Boolean(tokens.nextToken());
				this.data[i][j]=bol.booleanValue();
			}
		}
	}

	/**
	 * save the current Matrix
	 * @param fileName Filename
	 * @throws Exception
	 */

	public void save(String fileName) throws Exception {
		WriteFile writeFile=new WriteFile(fileName);
		writeFile.open();

		writeFile.writeLine(this.getDimensionX()+"\n");
		writeFile.writeLine(this.getDimensionY()+"\n");

		for (int i=0;i<this.data.length;i++) {
			for (int j=0;j<this.data[i].length;j++) {
				if (this.data[i][j])
					writeFile.writeLine("TRUE");
				else
					writeFile.writeLine("FALSE");

				if (j<(this.data[i].length-1))
					writeFile.writeLine(" ");
				else
					writeFile.writeLine("\n");
			}
		}

		writeFile.close();
	}

	/**
	 * prints the current Matrix
	 * @throws Exception
	 */

	public void print() throws Exception{
		for (int i=0; i<this.data.length; i++) {
			for (int j=0; j<this.data[i].length; j++) {
				if (this.data[i][j]==true)
					System.out.print("*");
				else
					System.out.print(" ");

				if (j<(this.data[i].length-1))
					System.out.print(" ");
				else
					System.out.println();
			}
		}
	}

	/**
	 * counts the neighbours of certain cell
	 * @param x width
	 * @param y height
	 * @param level depth of recursive count
	 * @return int
	 */

	public int countNeighbours(int x, int y, int level) {

		/**
		 * start in top left corner
		 */

		int countx=x-1;
		int county=y-1;

		int count=0;

		for (int i=0; i<3; i++) {
			if (!hasBounds) {
				if (countx<0)
					countx=this.data.length-1;
				if (countx>(this.data.length-1))
					countx=0;
		}

		for (int j=0;j<3;j++) {
			if (!hasBounds) {
				if (county<0)
					county=this.data[0].length-1;
				if (county>(this.data[0].length-1))
					county=0;
            	}

			/**
			 * don't count inner cell,
			 * only it's neigbours
			 */

			if (!((countx==x) && (county==y)) && countx>=0 && countx<data.length && county>=0 && county<data[0].length) {

					/**
					 * if neighbour is alive, count it
					 */

					if (this.data[countx][county]==true) {
						count++;

					/**
					 * recursive call
					 */

					if (level>0)
						count+=this.countNeighbours(countx, county, level-1);
					}
				}

				/**
				 * nextcolumn
				 */

				county++;
			}

			/**
			 * next line
			 */

			countx++;
			county=y-1;
		}

		return count;
	}

	/**
	 * nextGeneration -
	 * that's where the magic happens ;)
	 * @throws Exception
	 */

	public void nextGeneration() throws Exception {
		boolean oneAlive=false;

		/**
		 * create temporary Matrix
		 */

		boolean [][] newData=new boolean[this.data.length][this.data[0].length];

		for (int i=0;i<this.data.length;i++) {
			for (int j = 0; j < this.data[i].length; j++) {
				int neighbours=this.countNeighbours(i,j,0);

			/**
			 * ready to awake dead cell?
			 */

			if (this.data[i][j] == false) {
				if (neighbours == this.beBorn)
					newData[i][j] = true;
				else
					newData[i][j] = false;
			}

			/**
			 * keep cell alive?
			 */

			else {
				if (neighbours >= this.keepAlive_min && neighbours <= this.keepAlive_max)
					newData[i][j] = true;
				else
					newData[i][j] = false;
			}

			/**
			 * check if civilisation is dead
			 */

			if (newData[i][j]==true)
				oneAlive=true;
		}
	}

	if (!oneAlive)
		throw new Exception("I'm sorry! - Your civilisation is dead!");

		/**
		 * overwrite original array
		 */

		this.data=newData;
	}

	/**
	 * check coordinates for
	 * for living or dead creature
	 * @param x pos-x
	 * @param y pos-y
	 * @return data[x][y]
	 */

	public boolean getValue(int x, int y) {
		return this.data[x][y];
	}

	/**
	 * checks if a cell is alive
	 * @return number of creatures
	 */

	public int getAlive() {
		int count=0;

		for (int i=0;i<data.length;i++) {
			for (int j=0;j<data[i].length;j++) {
				if (data[i][j]==true)
					count++;
			}
		}

		return count;
	}

	/**
	 * change creature to dead cell
	 * and vice-versa;
	 * checks for boundaries
	 * @param x pos-x
	 * @param y pos-y
	 */

	public void changeValue(int x, int y) {
		if (x<0) x=0;
		if (y<0) y=0;

		if (x>(this.getDimensionX()-1))
			x=this.getDimensionX()-1;

		if (y>(this.getDimensionY()-1))
			y=this.getDimensionY()-1;
    	
		if (this.data[x][y])
			this.data[x][y]=false;
		else
			this.data[x][y]=true;
	}

	/**
	 * get width of Universe
	 * @return width
	 */

	public int getDimensionX() {
		return this.data.length;
	}

	/**
	 * get height of Universe
	 * @return height
	 */

	public int getDimensionY() {
		return this.data[0].length;
	}

	/**
	 * setter for hasBounds
	 */
	
	public void setBounds(boolean hasBounds) {
		this.hasBounds=hasBounds;
	}
}